#include "genesis.h"

#include "fmmusic.h"

/* very low level routines, don't compensate for channel, etc. */

/*
Sets a register in the FIRST FM chip
*/
void set_FM1(reg,dat)
unsigned char reg;
unsigned char dat;
{
   uchar *pFM;

   pFM = (uchar *) FM1_REG;

   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = reg;
   pFM++; /* advance from FMx_REG to FMx_DAT */
   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = dat;
}

/*
Sets a register in the SECOND FM chip
*/
void set_FM2(reg,dat)
unsigned char reg;
unsigned char dat;
{
   uchar *pFM;

   pFM = (uchar *) FM2_REG;

   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = reg;
   pFM++; /* advance from FMx_REG to FMx_DAT */
   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = dat;
}

/*
Higher level routine than the previous two.
Knows which channels are on which chip.
*/
void set_FM(chan,reg,dat)
unsigned char chan;
unsigned char reg;
unsigned char dat;
{
   uchar *pFM;

   /* First three channels go to chip FM1, second three go to chip FM2 */
   /* Registers 30H through B4H are "per voice channel", adjust */
   if (chan <= CHAN3) {
      pFM = (uchar *) FM1_REG;
      /* registers 30H through B4H are "per voice channel" */
      if ( (reg >= 0x30) && (reg <= 0xb4) )
         reg += chan-CHAN1;
   } else {
      pFM = (uchar *) FM2_REG;
      /* registers 30H through B4H are "per voice channel" */
      if ( (reg >= 0x30) && (reg <= 0xb4) )
         reg += chan-CHAN4;
   }
   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = reg;
   pFM++; /* advance from FMx_REG to FMx_DAT */
   /* wait until "chip not busy" */
   while (*pFM & 0x80)
      ;
   *pFM = dat;
}

/* just a convenience function */
void set_all_FM(reg,dat)
unsigned char reg;
unsigned char dat;
{
   set_FM(CHAN1,reg,dat);
   set_FM(CHAN2,reg,dat);
   set_FM(CHAN3,reg,dat);
   set_FM(CHAN4,reg,dat);
   set_FM(CHAN5,reg,dat);
   set_FM(CHAN6,reg,dat);
}


/* METHOD 2 OF GETTING VOICES */
/* These register settings are assumed by the following voice "patches" */
/* (These are reg,dat pairs with a 0xff terminator) */
uchar inits[] = {
0x21,0x00,0x24,0x00,0x25,0x00,0x2c,0x00,0x22,0x08,
0x27,0x00,0x2b,0x00,0xb4,0xc0,0xb5,0xc0,0xb6,0xc0,
0xb0,0x28,0xb1,0x28,0xb2,0x28,0xb0,0xf0,0x30,0x02,
0x34,0x71,0x38,0x00,0x3c,0x00,0x40,0x20,0x44,0x27,
0x48,0x23,0x4c,0x0f,0x50,0x1f,0x54,0x5f,0x58,0x5f,
0x5c,0x5f,0x60,0x0e,0x64,0x0f,0x68,0x07,0x6c,0x07,
0x70,0x07,0x74,0x00,0x78,0x00,0x7c,0x00,0x80,0xe3,
0x84,0xf4,0x88,0xf4,0x8c,0xf4,0x90,0x00,0x94,0x00,
0x98,0x00,0x9c,0x00,0xff};

char TVOICE[] = { /* a pretty good sounding "organ" type voice */
0x00,0x63,0x00,0x0B,0x1F,0x0F,0x0A,0x00,
0x01,0x63,0x00,0x0C,0x1F,0x0F,0x0C,0x00,
0x02,0x63,0x00,0x0B,0x1F,0x0F,0x0E,0x00,
0x03,0x63,0x00,0x0C,0x1F,0x0F,0x0F,0x00,
0x07,0x00};

char SILENCE[] = { /* a pretty good sounding "organ" type voice */
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00};

/*
You have to call this routine once if you are going to use METHOD 2
*/
void doinits()
{
   uchar *p1;
   uchar reg;
   uchar dat;

   p1=inits;

   while (*p1 != 0xff) {
      reg = *p1++;
      dat = *p1++;
      if ( (reg >= 0x30) || (reg <= 0xb4) )
         set_all_FM(reg,dat);
      else
         set_FM(reg,dat);
   }
}


void load_operator(chan,offset,patch)
uchar chan;
int offset;
uchar *patch;
{
   uchar v1,v2;
   set_FM(chan,0x30+offset,*patch++);
   v1 = *patch++;
   set_FM(chan,0x40+offset,104-v1);
   v1 = *patch++;
   v2 = *patch++;
   set_FM(chan,0x50+offset,(v1<<5)|v2);
   set_FM(chan,0x60+offset,*patch++);
   v1 = *patch++;
   set_FM(chan,0x70+offset,*patch++);
   v2 = *patch++;
   v1 = (~v1)<<4;
   set_FM(chan,0x80+offset,v1|v2);
}

void load_voice(chan,patch)
uchar chan;
uchar *patch;
{
   uchar v1,v2;
   load_operator(chan,0x00,patch);
   patch += 8;
   load_operator(chan,0x04,patch);
   patch += 8;
   load_operator(chan,0x08,patch);
   patch += 8;
   load_operator(chan,0x0c,patch);
   patch += 8;
   v1 = *patch++;
   v2 = *patch; /* no ++ because we are done */
   set_FM(chan,0xb0,v1|(v2<<3));
   set_FM(chan,0xb4,0xc0);
}
/* END OF METHOD 2 */

/* convenience function */
void load_voices(patch)
uchar *patch;
{
   load_voice(CHAN1,patch);
   load_voice(CHAN2,patch);
   load_voice(CHAN3,patch);
   load_voice(CHAN4,patch);
   load_voice(CHAN5,patch);
   load_voice(CHAN6,patch);
}

/*
This has to be called once, before the low level routines can be used.
*/
void init_synth()
{
   doinits();
   load_voices(SILENCE); /* just loading a default voice */
}

/*
Musical Note -> Frequency look up tables.
*/
char freql[]=
{/*0x00,*/0x40,0xa0,0x00,0x40,0xc0,0x20,0x80,
0xf0,0x40,0x80,0xc0,0x00,0x40,0xa0,0x00,
0x40,0xc0,0x20,0x80,0xf0,0x40,0x80,0xc0,
0x00,0x40,0xa0,0x00,0x40,0xc0,0x20,0x80,
0xf0,0x40,0x80,0xc0,0x00,0x40,0xa0,0x00,
0x40,0xc0,0x20,0x80,0xf0,0x40,0x80,0xc0,
0x00,0x40,0xa0,0x00,0x40,0xc0,0x20,0x80,
0xf0,0x40,0x80,0xc0,0x00,0x40,0xa0,0x00,
0x40,0xc0,0x20,0x80,0xf0,0x40,0x80,0xc0,
0x00};

char freqh[]=
{/*0x05,*/0x05,0x05,0x06,0x06,0x06,0x07,0x07,
0x07,0x0c,0x0c,0x0c,0x0d,0x0d,0x0d,0x0e,
0x0e,0x0e,0x0f,0x0f,0x0f,0x14,0x14,0x14,
0x15,0x15,0x15,0x16,0x16,0x16,0x17,0x17,
0x17,0x1c,0x1c,0x1c,0x1d,0x1d,0x1d,0x1e,
0x1e,0x1e,0x1f,0x1f,0x1f,0x24,0x24,0x24,
0x25,0x25,0x25,0x26,0x26,0x26,0x27,0x27,
0x27,0x2c,0x2c,0x2c,0x2d,0x2d,0x2d,0x2e,
0x2e,0x2e,0x2f,0x2f,0x2f,0x34,0x34,0x34,
0x35};

void note_on(chan,note)
unsigned char chan;
unsigned char note;
{
   set_FM(chan,0xa4,freqh[note]);
   set_FM(chan,0xa0,freql[note]);
   set_FM(CHAN1,0x28,0xf0+chan);
}

void freq_on(chan,freq)
unsigned char chan;
int freq;
{
   set_FM(chan,0xa4,freq>>8);
   set_FM(chan,0xa0,freq&0xff);
   set_FM(CHAN1,0x28,0xf0+chan);
}

void voice_on(chan)
unsigned char chan;
{
   set_FM(CHAN1,0x28,0xf0+chan);
}

void voice_off(chan)
unsigned char chan;
{
   set_FM(CHAN1,0x28,0x00+chan);
}

/*
The previous code was "low level" support. What follows is a higher level
"song engine" that makes use of the previous routines. Even if you don't
like this particular "song engine format", you should be able to reuse
the code above this point.
*/

/* pointer to the start of the song */
uint *song[CHANNELS+1]; /* one wasted array element due to channel "gap" */
/* pointer to positions within the song */
uint *mark1[CHANNELS+1]; /* same wasted space */
uint *mark2[CHANNELS+1]; /* same wasted space */
uint *mark3[CHANNELS+1]; /* same wasted space */
uint *mark4[CHANNELS+1]; /* same wasted space */
uint *mark5[CHANNELS+1]; /* same wasted space */
uint *mark6[CHANNELS+1]; /* same wasted space */
uint *mark7[CHANNELS+1]; /* same wasted space */
uint *mark8[CHANNELS+1]; /* same wasted space */

/* "return addresses" within the song */
/* These are used when doing a CALLMARKx */
uint *retmark1[CHANNELS+1];
uint *retmark2[CHANNELS+1];
uint *retmark3[CHANNELS+1];
uint *retmark4[CHANNELS+1];
uint *retmark5[CHANNELS+1];
uint *retmark6[CHANNELS+1];
uint *retmark7[CHANNELS+1];
uint *retmark8[CHANNELS+1];

/* "loop counters" for use with LOOP_MARK */
uchar loopcnt1[CHANNELS+1];
uchar loopcnt2[CHANNELS+1];
uchar loopcnt3[CHANNELS+1];
uchar loopcnt4[CHANNELS+1];
uchar loopcnt5[CHANNELS+1];
uchar loopcnt6[CHANNELS+1];
uchar loopcnt7[CHANNELS+1];
uchar loopcnt8[CHANNELS+1];

uchar delay[CHANNELS+1];
uchar transpose[CHANNELS+1];

/*
Runs the "song interpreter" on a given channel.
You'll probably use generate_music() (below) instead.
*/
void generate_channel(chan)
uchar chan;
{
   uchar note;

   while (1) { /* we might execute multiple commands */
      if (delay[chan]) {
         delay[chan]--;
         if (delay[chan]==0)
            voice_off(chan);
         else
            return;
      } else {
         note = *(song[chan]) & 0xff;
         if ( note < 128 ) { /* actual note, not a command */
            note_on(chan,note+transpose[chan]);
            delay[chan] = *(song[chan]) >> 8;
            song[chan]++;
            return;
         } else {
            switch (note) {
            case NOTE_NONE:
               voice_off(chan);
               delay[chan] = *(song[chan]) >> 8;
               song[chan]++;
               return;
            case NOTE_STOP:
               voice_off(chan);
               return;
            case SETTRANS:
               transpose[chan] = *(song[chan]) >> 8;
               song[chan]++;
            break;
            case TRANSUP:
               transpose[chan] += *(song[chan]) >> 8;
               song[chan]++;
            break;
            case TRANSDN:
               transpose[chan] -= *(song[chan]) >> 8;
               song[chan]++;
            break;
            case MARK_1:
               song[chan]++;
               mark1[chan] = song[chan];
            break;
            case MARK_2:
               song[chan]++;
               mark2[chan] = song[chan];
            break;
/* MARK3-8 need to be added sometime */
            case CALLMARK_1:
               if ( mark1[chan] != 0 ) {
                  song[chan]++;
                  retmark1[chan] = song[chan];
                  song[chan] = mark1[chan];
               } else {
                  song[chan]++;
               }
            break;
            case CALLMARK_2:
               if ( mark2[chan] != 0 ) {
                  song[chan]++;
                  retmark2[chan] = song[chan];
                  song[chan] = mark2[chan];
               } else {
                  song[chan]++;
               }
            break;
/* CALLMARK3-8 need to be added sometime */
            case GOMARK_1:
               if ( mark1[chan] != 0 ) {
                  song[chan] = mark1[chan];
               } else {
                  song[chan]++;
               }
            break;
            case GOMARK_2:
               if ( mark2[chan] != 0 ) {
                  song[chan] = mark2[chan];
               } else {
                  song[chan]++;
               }
            break;
/* GOMARK3-8 need to be added sometime */
            case ENDMARK_1:
               if ( retmark1[chan] != 0 ) {
                  song[chan] = retmark1[chan];
                  retmark1[chan] = 0;
               } else {
                  song[chan]++;
               }
            break;
            case ENDMARK_2:
               if ( retmark2[chan] != 0 ) {
                  song[chan] = retmark2[chan];
                  retmark2[chan] = 0;
               } else {
                  song[chan]++;
               }
            break;
/* ENDMARK3-8 need to be added sometime */
            }
         } /* end switch */
      } /* end else "not a note" */
   } /* end else "not in a delay" */
}

/*
You have to call this subroutine every 1/60 second.
(Less often will still run, but will throw off the note timings).
*/
void generate_music()
{
   generate_channel(CHAN1);
   generate_channel(CHAN2);
   generate_channel(CHAN3);
   generate_channel(CHAN4);
   generate_channel(CHAN5);
   generate_channel(CHAN6);
}

/*
This routine starts a song on a single channel.
Notice the trans (transposition) parameter.
Also realize that you can reuse song definitions
on multiple channels, with or without transposition.
(for example, octave parts)
*/
start_song(chan,trans,song_start)
uchar chan;
uint *song_start;
{
   transpose[chan] = trans;
   song[chan] = song_start;
}

uint parking_space = NOTE_STOP;

/*
You have to call this once, before you can use the song "engine".
*/
void init_song_player()
{
   uchar index;
   for (index=CHAN1; index<=CHAN6; index++) {
      delay[index] = 0;
      transpose[index] = 0;
      mark1[index] = 0;
      mark2[index] = 0;
      mark3[index] = 0;
      mark4[index] = 0;
      mark5[index] = 0;
      mark6[index] = 0;
      mark7[index] = 0;
      mark8[index] = 0;
      retmark1[index] = 0;
      retmark2[index] = 0;
      retmark3[index] = 0;
      retmark4[index] = 0;
      retmark5[index] = 0;
      retmark6[index] = 0;
      retmark7[index] = 0;
      retmark8[index] = 0;
      loopcnt1[index] = 0;
      loopcnt2[index] = 0;
      loopcnt3[index] = 0;
      loopcnt4[index] = 0;
      loopcnt5[index] = 0;
      loopcnt6[index] = 0;
      loopcnt7[index] = 0;
      loopcnt8[index] = 0;
      song[index] = &parking_space;
   }
}

