/* SEGA Genesis Memory Mapping */
#define FM1_REG   0xa04000
#define FM1_DAT   0xa04001
#define FM2_REG   0xa04002
#define FM2_DAT   0xa04003

/* MY chan numbering is 1-3, 4-6 */
/* HARDWARE channel numbers are 0-2, 4-6 */
/* ALWAYS use the following #defines */
#define CHANNELS   6

#define CHAN1   0
#define CHAN2   1
#define CHAN3   2
#define CHAN4   4
#define CHAN5   5
#define CHAN6   6

/* NOTE names. "S" is for "S"harp, "F" is for "F"lat. */
/* Both the "sharp" and the "flat" names are defined. */
/* The end digits are the octave, where 1 is lowest. */
/* 1 note lower is theoretically possible */
#define NOTE_C1   0
#define NOTE_CS1  1
#define NOTE_DF1  1
#define NOTE_D1   2
#define NOTE_DS1  3
#define NOTE_EF1  3
#define NOTE_E1   4
#define NOTE_F1   5
#define NOTE_FS1  6
#define NOTE_GF1  6
#define NOTE_G1   7
#define NOTE_GS1  8
#define NOTE_AF1  8
#define NOTE_A1   9
#define NOTE_AS1  10
#define NOTE_BF1  10
#define NOTE_B1   11
#define NOTE_C2   12
#define NOTE_CS2  13
#define NOTE_DF2  13
#define NOTE_D2   14
#define NOTE_DS2  15
#define NOTE_EF2  15
#define NOTE_E2   16
#define NOTE_F2   17
#define NOTE_FS2  18
#define NOTE_GF2  18
#define NOTE_G2   19
#define NOTE_GS2  20
#define NOTE_AF2  20
#define NOTE_A2   21
#define NOTE_AS2  22
#define NOTE_BF2  22
#define NOTE_B2   23
#define NOTE_C3   24
#define NOTE_CS3  25
#define NOTE_DF3  25
#define NOTE_D3   26
#define NOTE_DS3  27
#define NOTE_EF3  27
#define NOTE_E3   28
#define NOTE_F3   29
#define NOTE_FS3  30
#define NOTE_GF3  30
#define NOTE_G3   31
#define NOTE_GS3  32
#define NOTE_AF3  32
#define NOTE_A3   33
#define NOTE_AS3  34
#define NOTE_BF3  34
#define NOTE_B3   35
#define NOTE_C4   36
#define NOTE_CS4  37
#define NOTE_DF4  37
#define NOTE_D4   38
#define NOTE_DS4  39
#define NOTE_EF4  39
#define NOTE_E4   40
#define NOTE_F4   41
#define NOTE_FS4  42
#define NOTE_GF4  42
#define NOTE_G4   43
#define NOTE_GS4  44
#define NOTE_AF4  44
#define NOTE_A4   45
#define NOTE_AS4  46
#define NOTE_BF4  46
#define NOTE_B4   47
#define NOTE_C5   48
#define NOTE_CS5  48+1 /* funny numbering is from me using cut-n-paste */
#define NOTE_DF5  48+1 /* to make these octaves, sorry! */
#define NOTE_D5   48+2
#define NOTE_DS5  48+3
#define NOTE_EF5  48+3
#define NOTE_E5   48+4
#define NOTE_F5   48+5
#define NOTE_FS5  48+6
#define NOTE_GF5  48+6
#define NOTE_G5   48+7
#define NOTE_GS5  48+8
#define NOTE_AF5  48+8
#define NOTE_A5   48+9
#define NOTE_AS5  48+10
#define NOTE_BF5  48+10
#define NOTE_B5   48+11
#define NOTE_C6   60
#define NOTE_CS6  60+1
#define NOTE_DF6  60+1
#define NOTE_D6   60+2
#define NOTE_DS6  60+3
#define NOTE_EF6  60+3
#define NOTE_E6   60+4
#define NOTE_F6   60+5
#define NOTE_FS6  60+6
#define NOTE_GF6  60+6
#define NOTE_G6   60+7
#define NOTE_GS6  60+8
#define NOTE_AF6  60+8
#define NOTE_A6   60+9
#define NOTE_AS6  60+10
#define NOTE_BF6  60+10
#define NOTE_B6   60+11
#define NOTE_C7   60+0
#define NOTE_CS7  60+1
#define NOTE_DF7  60+1
#define NOTE_D7   60+2
#define NOTE_DS7  60+3
#define NOTE_EF7  60+3
#define NOTE_E7   60+4
#define NOTE_F7   60+5
#define NOTE_FS7  60+6
#define NOTE_GF7  60+6
#define NOTE_G7   60+7
#define NOTE_GS7  60+8
#define NOTE_AF7  60+8
#define NOTE_A7   60+9
#define NOTE_AS7  60+10
#define NOTE_BF7  60+10
#define NOTE_B7   60+11
#define NOTE_C8   72+0
#define NOTE_CS8  72+1
#define NOTE_DF8  72+1
#define NOTE_D8   72+2
#define NOTE_DS8  72+3
#define NOTE_EF8  72+3
#define NOTE_E8   72+4
#define NOTE_F8   72+5
#define NOTE_FS8  72+6
#define NOTE_GF8  72+6
#define NOTE_G8   72+7
#define NOTE_GS8  72+8
#define NOTE_AF8  72+8
#define NOTE_A8   72+9
#define NOTE_AS8  72+10
#define NOTE_BF8  72+10
#define NOTE_B8   72+11
#define NOTE_C9   84+0
#define NOTE_CS9  84+1
#define NOTE_DF9  84+1
#define NOTE_D9   84+2
#define NOTE_DS9  84+3
#define NOTE_EF9  84+3
#define NOTE_E9   84+4
#define NOTE_F9   84+5
#define NOTE_FS9  84+6
#define NOTE_GF9  84+6
#define NOTE_G9   84+7
#define NOTE_GS9  84+8
#define NOTE_AF9  84+8
#define NOTE_A9   84+9
#define NOTE_AS9  84+10
#define NOTE_BF9  84+10
#define NOTE_B9   84+11

/***********/
/* OPCODES */
/***********/

/* "MARK" a spot in the song so we can call it or jump to it later */
#define MARK_1  128
#define MARK_2  129
#define MARK_3  130
#define MARK_4  131
#define MARK_5  132
#define MARK_6  133
#define MARK_7  134
#define MARK_8  135

/* "MARK" the end of a section. Sort of like a RTS in 68K lingo */
#define ENDMARK_1  136
#define ENDMARK_2  137
#define ENDMARK_3  138
#define ENDMARK_4  139
#define ENDMARK_5  140
#define ENDMARK_6  141
#define ENDMARK_7  142
#define ENDMARK_8  143

/* "CALL" a previously defined (using MARKx/ENDMARKx) subroutine */
#define CALLMARK_1 144
#define CALLMARK_2 145
#define CALLMARK_3 146
#define CALLMARK_4 147
#define CALLMARK_5 148
#define CALLMARK_6 149
#define CALLMARK_7 150
#define CALLMARK_8 151

/* "GO" (jump) to a previous MARK */
#define GOMARK_1   152
#define GOMARK_2   153
#define GOMARK_3   154
#define GOMARK_4   155
#define GOMARK_5   156
#define GOMARK_6   157
#define GOMARK_7   158
#define GOMARK_8   159

/* for handling repeats */
#define LOOPMARK_1   160
#define LOOPMARK_2   161
#define LOOPMARK_3   162
#define LOOPMARK_4   163
#define LOOPMARK_5   164
#define LOOPMARK_6   165
#define LOOPMARK_7   166
#define LOOPMARK_8   167

#define SETTRANS   251 /* absolute transposition */
#define TRANSUP    252 /* relative transposition */
#define TRANSDN    253 /* "" */

#define NOTE_NONE 254
#define NOTE_STOP 255

/* NOTE LENGTHS. NOTICE how 8ths and 16ths require special care. */
#define WHOLE_NOTE   (60 << 8)
#define HALF_NOTE    (30 << 8)
#define QUARTER_NOTE (15 << 8)
#define FIRST_8TH    (8 << 8) /* these don't go in evenly */
#define SECOND_8TH   (7 << 8) /* these don't go in evenly */
#define TRIPLET_NOTE (5 << 8)
#define FIRST_16TH   (4 << 8) /* these don't go in evenly */
#define SECOND_16TH  (4 << 8) /* these don't go in evenly */
#define THIRD_16TH   (4 << 8) /* these don't go in evenly */
#define FOURTH_16TH  (3 << 8) /* these don't go in evenly */

/*
extern char TVOICE[];
extern uchar PIANO[];
extern uchar PLUKGUIT[];
extern char WINDS[];
extern char WAVE[];
extern char WILD_WAR[];
extern char XYLOPHON[];
extern char BOX_CELL[];
extern char STRINGS_[];
extern char TOM_TOMS[];
*/

/* See the .C file for block comments... */

/* Low level routines */

void set_FM1(reg,dat);
void set_FM2(reg,dat);

void set_FM(chan,reg,dat);
void set_all_FM(reg,dat);

void doinits();
void load_operator(chan,offset,patch);
void load_voice(chan,patch);
void load_voices(patch);

void init_synth();

void note_on(chan,note);
void freq_on(chan,freq);
void voice_on(chan);
void voice_off(chan);

/* high level routines */

void generate_channel(chan); /* usually called by generate_music() */
void generate_music();

start_song(chan,trans,song_start);
void init_song_player();

